#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A container for analysis.

Contains single hypothesis of single aspect of things.
"""

import re
from sys import stderr

from .formats.omor_formatter import OmorFormatter


class Analysis:
    """Contains a single analysis of a token.

    Analysis is a hypothesis of what token's some features may be:
    morphological analysis contains morphosyntactic readings and segmentation
    contains segment markers.
    """

    def __init__(self):
        """Create an empty analysis

        @param raw  analysis in string form
        @param weight  penalty weight for analysis
        """
        #: underlying raw omor analysis
        self.raw = None
        #: universal part of speech
        self.upos = None
        #: last effective universal feats (private feats allowed, but see misc)
        self.ufeats = dict()
        #: UD dep, target and name
        self.udepname = None
        self.udeppos = None
        #: misc features
        self.misc = dict()
        #: type of analysis: omor, segmentation, lemmatisation etc.
        self.rawtype = None
        #: FSA or similar weight
        self.weight = float("inf")
        #: surface actually used for analysis (if different from parent token's
        #: surf)
        self.analsurf = None
        #: Recasers, spelling correctors and plain guessers
        self.manglers = list()
        #: lemmas
        self.lemmas = list()

    def __str__(self):
        s = '"omorfi.Analysis": {'
        if self.raw:
            s += '"rawtype": "' + self.rawtype + '", '
            s += '"raw": "' + self.raw + '", '
        if self.upos:
            s += '"upos": "' + self.upos + '", '
        if self.ufeats:
            s += '"ufeats": "' + str(self.ufeats) + '", '
        if self.misc:
            s += '"misc": "' + str(self.misc) + '", '
        s += '"weight": "' + str(self.weight) + '"'
        if self.udepname:
            s += '"udep": "' + self.udepname + '"'
            s += '"udephead": "' + str(self.udeppos) + '"'
        s += '}'
        return s

    def get_upos(self):
        '''Finds UPOS from analyses.

        Returns:
            upos in a string
        '''
        return self.upos

    def get_lemmas(self):
        '''Finds lemmas from analyses.

        Returns:
            list of strings.
        '''
        return self.lemmas

    def get_ufeats(self):
        '''Finds UD Feats from analyses.

        Returns:
            dict of key value pairs of UD Feat column.
        '''
        return self.ufeats

    @staticmethod
    def fromstr(s: str):
        '''Constructs analysis from string'''
        a = Analysis()
        if '"omorfi.Analysis":' not in s:
            return None
        start = s.find('"omorfi.Analysis":') + len('"omorfi.Analysis":')
        start = s.find("{", start)
        end = s.rfind("}")
        fields = s[start:end].split(", ")
        for field in fields:
            k, v = field.split(":")
            if k.strip('"') == "rawtype":
                a.rawtype = v.strip('"')
            elif k.strip('"') == "raw":
                a.raw = v.strip('"')
            elif k.strip('"') == "weight":
                a.weight = float(v.strip('"'))
            else:
                print("Cannot parse", s, "as analysis", file=stderr)
                exit(1)
        return a

    @staticmethod
    def fromomor(s: str, weight=float("inf"), hacks=None):
        '''Constructs analysis form Omor style string.

        Typically used to create an analysis from libhfst string and weight
        after using omorfi HFST analyser on a surface string.

        Args:
            s       An omor-style analysis string, e.g.
                "[WORD_ID=äh][UPOS=INTJ]"
            weight  A penalty-weight of the analysis
            hacks   Used for mangling some values based on some standards and
                treebanks

        Returns:
            a token with omor analysis parsed into structured information
        '''
        a = Analysis()
        a.raw = s
        a.rawtype = "omor"
        a.weight = weight
        a.upos = OmorFormatter.get_upos(s)
        a.lemmas = OmorFormatter.get_lemmas(s)
        a.weight += (len(a.lemmas) - 1) * 10
        a.weight += s.count('DRV') * 5
        feats = OmorFormatter.get_last_feats(s)
        if not feats:
            a.ufeats = None
            return a
        a.ufeats = dict()
        for f in feats:
            key = f.split("=")[0].lstrip("[")
            value = f.split("=")[1].rstrip("]")
            if key == 'CASE':
                if value == 'LAT' and hacks != 'ftb':
                    # XXX: hack to retain compability
                    a.ufeats['Number'] = 'Sing'
                else:
                    a.ufeats['Case'] = value[0] + value[1:].lower()
            elif key == 'NUM':
                if value == 'SG':
                    a.ufeats['Number'] = 'Sing'
                elif value == 'PL':
                    a.ufeats['Number'] = 'Plur'
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
            elif key == 'TENSE':
                if 'PRESENT' in value:
                    a.ufeats['Tense'] = 'Pres'
                elif 'PAST' in value:
                    a.ufeats['Tense'] = 'Past'
            elif key == 'MOOD':
                a.ufeats['VerbForm'] = 'Fin'
                if value == 'INDV':
                    a.ufeats['Mood'] = 'Ind'
                elif value == 'COND':
                    a.ufeats['Mood'] = 'Cnd'
                elif value == 'IMPV':
                    a.ufeats['Mood'] = 'Imp'
                elif value == 'POTN':
                    a.ufeats['Mood'] = 'Pot'
                elif value == 'OPT':
                    a.misc['Mood'] = 'Opt'
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
            elif key == 'VOICE':
                if value == 'PSS':
                    a.ufeats['Voice'] = 'Pass'
                elif value == 'ACT':
                    a.ufeats['Voice'] = 'Act'
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
            elif key == 'PERS':
                if 'SG' in value:
                    a.ufeats['Number'] = 'Sing'
                elif 'PL' in value:
                    a.ufeats['Number'] = 'Plur'
                elif 'PE' in value:
                    pass
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
                if '1' in value:
                    a.ufeats['Person'] = '1'
                elif '2' in value:
                    a.ufeats['Person'] = '2'
                elif '3' in value:
                    a.ufeats['Person'] = '3'
                elif '0' in value:
                    a.ufeats['Person'] = '0'
                elif '4' in value:
                    a.misc['Person'] = '4'
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
            elif key == 'POSS':
                if 'SG' in value:
                    a.ufeats['Number[psor]'] = 'Sing'
                elif 'PL' in value:
                    a.ufeats['Number[psor]'] = 'Plur'
                elif value == '3':
                    pass
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
                if '1' in value:
                    a.ufeats['Person[psor]'] = '1'
                elif '2' in value:
                    a.ufeats['Person[psor]'] = '2'
                elif '3' in value:
                    a.ufeats['Person[psor]'] = '3'
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
            elif key == 'NEG':
                if value == 'CON':
                    a.ufeats['Connegative'] = 'Yes'
                    # XXX
                    a.ufeats.pop('Voice')
                elif value == 'NEG':
                    a.ufeats['Polarity'] = 'Neg'
                    a.ufeats['VerbForm'] = 'Fin'
            elif key == 'PCP':
                a.ufeats['VerbForm'] = 'Part'
                if value == 'VA':
                    a.ufeats['PartForm'] = 'Pres'
                elif value == 'NUT':
                    a.ufeats['PartForm'] = 'Past'
                elif value in ["MA", "AGENT"]:
                    a.ufeats['PartForm'] = 'Agt'
                elif value in ["MATON", "NEG"]:
                    a.ufeats['PartForm'] = 'Neg'
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
            elif key == 'INF':
                a.ufeats['VerbForm'] = 'Inf'
                if value == 'A':
                    a.ufeats['InfForm'] = '1'
                elif value == 'E':
                    a.ufeats['InfForm'] = '2'
                    # XXX
                    a.ufeats['Number'] = 'Sing'
                elif value == 'MA':
                    a.ufeats['InfForm'] = '3'
                    # XXX
                    a.ufeats['Number'] = 'Sing'
                elif value == 'MINEN':
                    a.ufeats['Derivation'] = 'Minen'
                elif value == 'MAISILLA':
                    a.ufeats['InfForm'] = '5'
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
            elif key == 'CMP':
                if value == 'SUP':
                    a.ufeats['Degree'] = 'Sup'
                elif value == 'CMP':
                    a.ufeats['Degree'] = 'Cmp'
                elif value == 'POS':
                    a.ufeats['Degree'] = 'Pos'
                else:
                    print(key, value, 'omor', file=stderr)
                    exit(1)
            elif key == 'SUBCAT':
                if value == 'NEG':
                    a.ufeats['Polarity'] = 'Neg'
                    a.ufeats['VerbForm'] = 'Fin'
                elif value == 'QUANTIFIER':
                    a.ufeats['PronType'] = 'Ind'
                elif value == 'REFLEXIVE':
                    a.ufeats['Reflex'] = 'Yes'
                elif value in ['COMMA', 'DASH', 'QUOTATION', 'BRACKET',
                               'ARROW']:
                    # not annotated in UD feats:
                    # * punctuation classes
                    a.misc['PunctType'] = value[0] + value[1:].lower()
                elif value in ['DECIMAL', 'ROMAN', 'DIGIT']:
                    # not annotated in UD feats:
                    # * decimal, roman NumType
                    a.misc['NumType'] = value[0] + value[1:].lower()
                elif value in ['PREFIX', 'SUFFIX']:
                    # not annotated in UD feats:
                    # * Hanging prefix/suffix?
                    a.misc['AffixType'] = value[0] + value[1:].lower()
                else:
                    print(key, value, 'SUBCAT', 'UD')
                    exit(1)
            elif key == 'ABBR':
                # XXX?
                a.ufeats['Abbr'] = 'Yes'
            elif key == 'NUMTYPE':
                a.ufeats['NumType'] = value[0] + value[1:].lower()
            elif key == 'PRONTYPE':
                a.ufeats['PronType'] = value[0] + value[1:].lower()
            elif key == 'ADPTYPE':
                a.ufeats['AdpType'] = value[0] + value[1:].lower()
            elif key == 'CLIT':
                a.ufeats['Clitic'] = value[0] + value[1:].lower()
            elif key == 'FOREIGN':
                a.ufeats['Foreign'] = 'Yes'
            elif key == 'STYLE':
                if value in ['DIALECTAL', 'COLLOQUIAL']:
                    a.ufeats['Style'] = 'Coll'
                elif value == 'NONSTANDARD':
                    # XXX: Non-standard spelling is kind of a typo?
                    # e.g. seitsämän -> seitsemän
                    a.ufeats['Typo'] = 'Yes'
                elif value == 'ARCHAIC':
                    a.ufeats['Style'] = 'Arch'
                elif value == 'RARE':
                    continue
                else:
                    print(key, value, 'STYLE', 'UD')
                    exit(1)
            elif key == 'DRV':
                if value in ['INEN', 'JA', 'LAINEN', 'LLINEN', 'MINEN', 'STI',
                             'TAR', 'TON', 'TTAA', 'TTAIN', 'U', 'VS']:
                    # values found in UD finnish Derivs
                    a.ufeats['Derivation'] = value[0] + value[1:].lower()
                elif value in ['S', 'MAISILLA', 'VA', 'MATON', 'UUS', 'ADE',
                               'INE', 'ELA', 'ILL', 'NEN', 'MPI', 'IN', 'IN²',
                               'HKO', 'ISA', 'MAINEN', 'NUT', 'TU', 'VA',
                               'TAVA', 'MA', 'LOC', 'LA', 'TUTTAA']:
                    # valuse not found in UD finnish Derivs
                    a.misc['Derivation'] = value[0] + value[1:].lower()
                else:
                    print(key, value, 'DRV', 'UD')
                    exit(1)
            elif key == 'LEX':
                a.misc['Lexicalised'] = value[0] + value[1:].lower()
            elif key == 'BLACKLIST':
                a.misc['Blacklisted'] = value
            elif key == 'PROPER':
                a.misc['PropnType'] = value[0] + value[1:].lower()
            elif key == 'POSITION':
                a.misc['Position'] = value[0] + value[1:].lower()
            elif key == 'SEM':
                a.misc['SemType'] = value[0] + value[1:].lower()
            elif key == 'COMPOUND_FORM':
                a.misc['GoesWith'] = value[0] + value[1:].lower()
            elif key in ['UPOS', 'ALLO', 'WEIGHT', 'CASECHANGE', 'NEWPARA',
                         'GUESS', 'CONJ', 'BOUNDARY']:
                # Not feats in UD:
                # * UPOS is another field
                # * Allomorphy is ignored
                # * Weight = no pties
                # * No feats for recasing
                # * FIXME: lexicalised inflection usually not a feat
                # * Guessering not a feat
                # * Proper noun classification not a feat
                # * punct sidedness is not a feat
                # * XXX: sem has not been used as a feat?
                # * special CONJ comparative is not used in UD
                # * clause / sentence boundary tag ignored
                continue
            else:
                print(key, value, 'omor')
                exit(1)
        return a

    @staticmethod
    def fromvislcg(s):
        '''Constructs analysis from VISL-CG string.

        The string should match what the method printable_vislcg creates plus
        optional VISL CG 3 trace and such markings.
        '''
        fields = s.strip().split()
        lemma = fields[0]
        if not fields[0].startswith('"'):
            print("Cannot find lemma in VISL CG 3:", s, file=stderr)
            exit(1)
        if not fields[0].endswith('"'):
            lemmaend = False
            for posm1, field in enumerate(fields[1:]):
                lemma += field
                if field.endswith('"'):
                    lemmaend = posm1 + 1
                    break
            if not lemmaend:
                print("Cannot find lemma in VISL CG 3:", s, file=stderr)
                exit(1)
            fields = [lemma] + fields[lemmaend + 1:]
        lemma = fields[0].strip('"')
        vislcgs = []
        omorstr = '[WORD_ID=' + lemma + ']'
        weight = 0.0
        if len(fields) < 2:
            return omorstr
        upos = fields[1]
        if upos in ["NOUN", "ADJ", "VERB", "AUX", "ADP", "X", "CCONJ", "PRON",
                    "SCONJ", "DET", "PROPN", "PUNCT", "SYM", "ADV", "NUM",
                    "INTJ"]:
            omorstr += '[UPOS=' + upos + ']'
        else:
            print("Cannot find UPOS in VISL CG 3:", fields, file=stderr)
            exit(1)
        if len(fields) < 3:
            return omorstr
        vislcgs = fields[2:]
        for vislcg in vislcgs:
            if vislcg in ["PRESENT", "PAST"]:
                omorstr += '[TENSE=' + vislcg + ']'
            elif vislcg in ["SG", "PL"]:
                omorstr += '[NUM=' + vislcg + ']'
            elif vislcg in ["ABE", "ABL", "ACC", "ADE", "ALL", "COM", "ELA",
                            "ESS", "GEN", "ILL", "INE", "INS", "LAT", "NOM",
                            "PAR", "TRA"]:
                omorstr += "[CASE=" + vislcg + "]"
            elif vislcg in ["INDV", "IMPV", "COND", "POTN", "IND"]:
                omorstr += "[MOOD=" + vislcg + "]"
            elif vislcg in ["SG1", "SG2", "SG3", "PL1", "PL2", "PL3", "PE4"]:
                omorstr += "[PERS=" + vislcg + "]"
            elif vislcg in ["CMP", "POS", "SUP"]:
                omorstr += "[CMP=" + vislcg + "]"
            elif vislcg == "CONNEG":
                omorstr += "[NEG=CON]"
            elif vislcg in ["PRS", "DEM", "REL", "INT", "REC"]:
                omorstr += "[PRONTYPE=" + vislcg + "]"
            elif vislcg in ["HAN", "KO", "PA", "S", "KA", "KIN", "KAAN"]:
                omorstr += "[CLIT=" + vislcg + "]"
            elif vislcg in ["PREP", "POST"]:
                omorstr += "[ADPTYPE=" + vislcg + "]"
            elif vislcg in ["CARD", "ORD", "FRAC", "MULT"]:
                omorstr += "[NUMTYPE=" + vislcg + "]"
            elif vislcg in ["ACT", "PSS"]:
                omorstr += "[VOICE=" + vislcg + "]"
            elif vislcg in ["SCONJ", "COMPARATIVE", "ADP", "SGNOM"]:
                # XXX: hacks that shouldn't be
                pass
            elif vislcg.startswith("POSS"):
                omorstr += "[POSS=" + vislcg[4:] + "]"
            elif vislcg.startswith("CLIT"):
                omorstr += "[CLIT=" + vislcg[4:] + "]"
            elif vislcg.startswith("PCP"):
                omorstr += "[PCP=" + vislcg[3:] + "]"
            elif vislcg.startswith("INF"):
                omorstr += "[INF=" + vislcg[3:] + "]"
            elif vislcg.startswith("<W=") and vislcg.endswith(">"):
                weight = float(vislcg[3:-1])
            elif vislcg.startswith("<") and vislcg.endswith(">"):
                # secondaries are safe to skip
                pass
            else:
                print("Cannot parse", vislcg, "as vislcg", file=stderr)
                exit(1)
        anal = Analysis.fromomor(omorstr)
        anal.weight = weight
        return anal

    def get_ftb_feats(self):
        '''Get ftb analyses from token data.'''
        feats = self.ufeats
        rvs = list()
        rvs += [self.get_xpos_ftb()]
        if self.upos == 'PROPN':
            rvs += ['Prop']
        elif self.upos == 'ADV' and self.get_lemmas()[-1].endswith("sti"):
            # This is FTB oddity
            rvs += ['Pos', 'Man']
        for key, value in feats.items():
            if key == 'Number':
                if value == 'Sing':
                    rvs += ['Sg']
                elif value == 'Plur':
                    rvs += ['Pl']
            elif key == 'Tense':
                if value == 'Pres':
                    rvs += ['Prs']
                elif value == 'Past':
                    rvs += ['Prt']
            elif key == 'Mood':
                if value == 'Ind':
                    continue
                elif value == 'Cnd':
                    rvs += ['Cond']
                elif value == 'Impv':
                    rvs += ['Imp']
                else:
                    rvs += [value]
            elif key == 'Voice':
                rvs += [value]
            elif key == 'Person':
                if value == '0':
                    rvs += ['__3']
                elif value == '1':
                    rvs += ['__1']
                elif value == '2':
                    rvs += ['__2']
                elif value == '3':
                    rvs += ['__3']
                elif value == '4':
                    rvs += ['Pe4']
                else:
                    print(key, value, "for ftb", file=stderr)
                    exit(1)
            elif key == 'Number[psor]':
                if value == 'Sing':
                    rvs += ['PxSg_']
                elif value == 'Plur':
                    rvs += ['PxPl_']
                else:
                    print(key, value, "for ftb", file=stderr)
                    exit(1)
            elif key == 'Person[psor]':
                if value == '1':
                    rvs += ['Px__1']
                elif value == '2':
                    rvs += ['Px__2']
                elif value == '3':
                    rvs += ['PxSp3']
                else:
                    print(key, value, "for ftb", file=stderr)
                    exit(1)
            elif key == 'Polarity':
                if value == 'Neg':
                    rvs += ['Neg']
                else:
                    print(key, value, "ftb", file=stderr)
                    exit(1)
            elif key == 'Connegative':
                if value == 'Yes':
                    rvs += ['Act', 'ConNeg']
                else:
                    print(key, value, "ftb", file=stderr)
                    exit(1)
            elif key == 'InfForm':
                if value == '1':
                    rvs += ['Inf1', 'Lat']
                elif value == '2':
                    rvs += ['Inf2']
                elif value == '3':
                    rvs += ['Inf3']
                elif value == 'MINEN':
                    rvs += ['Inf4']
                elif value == 'MAISILLA':
                    rvs += ['Inf5']
            elif key == 'PartForm':
                # FTB participle is POS
                pass
            elif key == 'Case':
                rvs += [value]
            elif key == 'Degree':
                rvs += [value]
            elif key == 'SUBCAT':
                if value == 'NEG':
                    rvs += ['Neg']
                elif value == 'QUOTATION':
                    rvs += ['Quote']
                elif value == 'QUANTIFIER':
                    rvs += ['Qnt']
                elif value == 'DIGIT':
                    rvs += ['Digit']
                elif value in ['COMMA', 'BRACKET',
                               'ARROW', 'DECIMAL', 'PREFIX', 'SUFFIX']:
                    # not annotated in FTN feats:
                    # * punctuation classes
                    continue
                elif value == 'ROMAN':
                    # not annotated in FTN feats:
                    # * decimal, roman NumType
                    continue
                else:
                    print(key, value, 'SUBCAT', 'FTB3')
                    exit(1)
            elif key == 'NumType':
                if value == 'Ord':
                    rvs += [value]
                else:
                    pass
            elif key == 'PronType':
                if value == 'Prs':
                    rvs += ['Pers']
                elif value == 'Ind':
                    rvs += ['Qnt']
                elif value == 'Int':
                    rvs += ['Interr']
                else:
                    rvs += [value]
            elif key == 'AdpType':
                if value == 'Post':
                    rvs += ['Po']
                elif value == 'Prep':
                    rvs += ['Pr']
                else:
                    print(key, value, 'ADPTYPE', 'FTB3')
                    exit(1)
            elif key == 'Clitic':
                if value == 'Ka':
                    rvs += ['Foc_kA']
                else:
                    rvs += ['Foc_' + value]
            elif key == 'Abbr':
                rvs += ['Abbr']
            elif key == 'Derivation':
                if value in ['NUT', 'VA']:
                    rvs += ['Act']
                elif value in ['TU', 'TAVA']:
                    rvs += ['Pass']
                else:
                    continue
            elif key == 'Reflex':
                rvs += ['Refl']
            elif key in ['UPOS', 'ALLO', 'WEIGHT', 'CASECHANGE', 'NEWPARA',
                         'GUESS', 'PROPER', 'SEM', 'CONJ', 'BOUNDARY',
                         'PCP', 'DRV', 'LEX', 'BLACKLIST', 'Style',
                         'POSITION', "Foreign", 'VerbForm',
                         'Typo']:
                continue
            else:
                print(key, value, 'FTB3')
                exit(1)
        for key, value in self.misc.items():
            if key == 'NumType':
                rvs += [value]
            elif key == 'Person' and value == '4':
                rvs += ['Pe4']
            elif key == 'PunctType':
                if value == "Quotation":
                    rvs += ["Quote"]
                elif value == "Dash":
                    if self.get_lemmas()[-1] == '—':
                        rvs += ['EmDash']
                    elif self.get_lemmas()[-1] == '–':
                        rvs += ['EnDash']
                    else:
                        rvs += ['Dash']
                elif value in ["Comma", "Bracket", "Arrow"]:
                    pass
                else:
                    print(key, value, 'FTB3', file=stderr)
                    exit(1)
            elif key == 'PropnType':
                rvs += ["Prop"]
            elif key in ['AffixType', "GoesWith", "Position"]:
                # XXX
                pass
            elif key == 'SemType':
                pass
            elif key == "Derivation":
                if value in ["Tava", "Tu"]:
                    rvs += ["Pass"]
                elif value in ["Va", "Nut"]:
                    rvs += ["Act"]
                else:
                    pass
            elif key == "Mood":
                if value == 'Opt':
                    rvs += ["Opt"]
            elif key in ["Lexicalised", "Blacklisted"]:
                continue
            else:
                print(key, value, 'FTB3miscelse', file=stderr)
                exit(1)
        # post hacks
        if self.lemmas == ['ei'] and 'Foc_kA' in rvs:
            revs = []
            for r in rvs:
                if r != 'V':
                    revs += [r]
                else:
                    revs += ['CC']
            rvs = revs
        if 'Punct' in rvs and 'Sg' in rvs and 'Nom' in rvs:
            revs = []
            for r in rvs:
                if r not in ['Sg', 'Nom']:
                    revs += [r]
            rvs = revs
        if '__1' in rvs or '__2' in rvs or '__3' in rvs:
            revs = []
            for r in rvs:
                if r not in ['__1', '__2', '__3', 'Sg', 'Pl']:
                    revs += [r]
            if 'Sg' in rvs and '__1' in rvs:
                revs += ['Sg1']
            elif 'Sg' in rvs and '__2' in rvs:
                revs += ['Sg2']
            elif 'Sg' in rvs and '__3' in rvs:
                revs += ['Sg3']
            elif 'Pl' in rvs and '__1' in rvs:
                revs += ['Pl1']
            elif 'Pl' in rvs and '__2' in rvs:
                revs += ['Pl2']
            elif 'Pl' in rvs and '__3' in rvs:
                revs += ['Pl3']
            else:
                print("__X without Sg or Pl", file=stderr)
            rvs = revs
        if 'Px__1' in rvs or 'Px__2' in rvs or 'Px__3' in rvs:
            revs = []
            for r in rvs:
                if r not in ['Px__1', 'Px__2', 'Px__3', 'PxSg_', 'PxPl_']:
                    revs += [r]
            if 'PxSg_' in rvs and 'Px__1' in rvs:
                revs += ['PxSg1']
            elif 'PxSg_' in rvs and 'Px__2' in rvs:
                revs += ['PxSg2']
            elif 'PxSg_' in rvs and 'Px__3' in rvs:
                revs += ['PxSg3']
            elif 'PxPl_' in rvs and 'Px__1' in rvs:
                revs += ['PxPl1']
            elif 'PxPl_' in rvs and 'Px__2' in rvs:
                revs += ['PxPl2']
            elif 'PxPl_' in rvs and 'Px__3' in rvs:
                revs += ['PxPl3']
            elif 'Px__3' in rvs:
                revs += ['Px3']
            else:
                print("__X without Sg or Pl", file=stderr)
            rvs = revs
        if 'Neg' in rvs and 'Act' in rvs:
            revs = []
            for r in rvs:
                if r != 'Act':
                    revs += [r]
            rvs = revs
        if 'Abbr' in rvs:
            revs = []
            for r in rvs:
                if r not in ['N', 'Prop']:
                    revs += [r]
            rvs = revs
        if 'Inf1' in rvs:
            revs = []
            for r in rvs:
                if r not in ['Act', 'Pl', 'Sg']:
                    revs += [r]
            rvs = revs
        if 'Pers' in rvs:
            revs = []
            for r in rvs:
                if r not in ['Pl1', 'Sg1', 'Pl2', 'Sg2', 'Pl3', 'Sg3']:
                    revs += [r]
                elif r in ['Pl1', 'Pl2', 'Pl3']:
                    revs += ['Pl']
                elif r in ['Sg1', 'Sg2', 'Sg3']:
                    revs += ['Sg']
                else:
                    print("Logical error", r, rvs)
                    exit(2)
            rvs = revs
        if 'Card' in rvs and 'Digit' in rvs:
            revs = []
            for r in rvs:
                if r != 'Card':
                    revs += [r]
            rvs = revs
        return rvs

    def get_unimorph_feats(self):
        '''Get Unimorph analyses from token data.'''
        rvs = list()
        if self.upos == 'NOUN':
            rvs += ['N']
        elif self.upos == 'VERB':
            if 'PartForm' in self.ufeats:
                rvs += ['V.PTCP']
            else:
                rvs += ['V']
        elif self.upos == 'ADJ':
            rvs += ['ADJ']
        for key, value in self.ufeats.items():
            if key == 'Number':
                if value == 'Sing':
                    rvs += ['SG']
                elif value == 'Plur':
                    rvs += ['PL']
            elif key == 'Tense':
                if value == 'Pres':
                    rvs += ['PRS']
                elif value == 'Past':
                    rvs += ['PST']
            elif key == 'Mood':
                if value == 'Ind':
                    rvs += ['POS', 'IND']
                elif value == 'Cnd':
                    rvs += ['PRS', 'POS', 'COND']
                elif value == 'Imp':
                    rvs += ['PRS', 'POS', 'IMP']
                elif value == 'Pot':
                    rvs += ['PRS', 'POS', 'POT']
                else:
                    rvs += [value.upper()]
            elif key == 'Voice':
                rvs += [value.upper()]
            elif key == 'Person':
                if value == '0':
                    rvs += ['3']
                elif value == '1':
                    rvs += ['1']
                elif value == '2':
                    rvs += ['2']
                elif value == '3':
                    rvs += ['3']
                elif value == '4':
                    pass
                else:
                    print(key, value, "for unimorph", file=stderr)
                    exit(1)
            elif key == 'Number[psor]':
                # ignored in unimroph
                pass
            elif key == 'Person[psor]':
                # ignored in unimroph
                pass
            elif key == 'Polarity':
                rvs += ['???']
            elif key == 'Connegative':
                rvs += ['NEG']
            elif key == 'InfForm':
                if value == '1':
                    rvs += ['NFIN']
                elif value == '2':
                    rvs += ['???']
                elif value == '3':
                    rvs += ['???']
                elif value == 'MINEN':
                    rvs += ['???']
                elif value == 'MAISILLA':
                    rvs += ['???']
            elif key == 'PartForm':
                if value == 'Pres':
                    rvs += ['PRS']
                elif value == 'Past':
                    rvs += ['PST']
                elif value == '3':
                    rvs += ['???']
            elif key == 'Case':
                if value == 'Nom':
                    rvs += ['NOM']
                elif value == 'Ill':
                    rvs += ['IN+ALL']
                elif value == 'Tra':
                    rvs += ['TRANS']
                elif value == 'Ade':
                    rvs += ['AT+ESS']
                elif value == 'All':
                    rvs += ['AT+ALL']
                elif value == 'Abl':
                    rvs += ['AT+ABL']
                elif value == 'Ess':
                    rvs += ['FRML']
                elif value == 'Com':
                    rvs += ['COM', 'PL']
                elif value == 'Ins':
                    rvs += ['INS']
                elif value == 'Ine':
                    rvs += ['IN+ESS']
                elif value == 'Ela':
                    rvs += ['IN+ABL']
                elif value == 'Abe':
                    rvs += ['PRIV']
                elif value == 'Par':
                    rvs += ['PRT']
                elif value == 'Gen':
                    rvs += ['GEN']
                elif value == 'Acc':
                    rvs += ['ACC']
                else:
                    rvs += ['???']
            elif key == 'Degree':
                # yeah. unimorph has same analysis for comparative and positive
                pass
            elif key == 'SUBCAT':
                rvs += ['???']
            elif key == 'NumType':
                rvs += ['???']
            elif key == 'PronType':
                rvs += ['???']
            elif key == 'AdpType':
                rvs += ['???']
            elif key == 'Clitic':
                rvs += ['???']
            elif key == 'Abbr':
                rvs += ['???']
            elif key == 'Derivation':
                rvs += ['???']
            elif key == 'Reflex':
                rvs += ['???']
            elif key in ['UPOS', 'ALLO', 'WEIGHT', 'CASECHANGE', 'NEWPARA',
                         'GUESS', 'PROPER', 'SEM', 'CONJ', 'BOUNDARY',
                         'PCP', 'DRV', 'LEX', 'BLACKLIST', 'Style',
                         'POSITION', "Foreign", 'VerbForm',
                         'Typo']:
                continue
            else:
                print(key, value, 'unimorph')
                exit(1)
        # unimorph is very limited so we delete a lot of tags...
        if 'V.PTCP' in rvs and 'ACT' in rvs and 'PRS' in rvs:
            rvs = ['V.PTCP', 'ACT', 'PRS']
        elif 'V.PTCP' in rvs and 'ACT' in rvs and 'PST' in rvs:
            rvs = ['V.PTCP', 'ACT', 'PST']
        elif 'V.PTCP' in rvs and 'PASS' in rvs and 'PRS' in rvs:
            rvs = ['V.PTCP', 'PASS', 'PRS']
        elif 'V.PTCP' in rvs and 'PASS' in rvs and 'PST' in rvs:
            rvs = ['V.PTCP', 'PASS', 'PST']
        elif 'NFIN' in rvs:
            rvs = ['V', 'NFIN']
        # re-order as use
        revs = []
        for r in rvs:
            if r not in ['SG', 'PL']:
                revs.append(r)
        for r in rvs:
            if r in ['SG', 'PL']:
                revs.append(r)
        return revs

    def get_vislcg_feats(self):
        '''Get VISL-CG 3 features from analysed token.'''
        vislcgs = list()
        vislcgs += ['UPOS=' + self.upos]
        for key, value in self.ufeats.items():
            vislcgs += [key + '=' + value]
        for key, value in self.misc.items():
            vislcgs += ['<' + key + '=' + value + '>']
        # number of compound parts in compound is a good CG numeric feature!!
        lemmas = self.get_lemmas()
        vislcgs += ['<CMP=' + str(len(lemmas)) + '>']
        if self.weight != float('inf'):
            vislcgs += ['<W=' + str(int(self.weight * 1000)) + '>']
        else:
            vislcgs += ['<W=999999999999999999999999>']
        for mangler in self.manglers:
            vislcgs += ['<*' + mangler + '>']
        return vislcgs

    def get_segments(self, split_morphs=True, split_words=True,
                     split_new_words=True, split_derivs=False,
                     split_nonwords=False):
        '''Get specified segments from segmented analysis.'''
        if self.rawtype != 'segments':
            return None
        segments = self.raw
        # this code is ugly
        segments = [segments.replace('{hyph?}', '').replace("{STUB}", "")]
        resegs = []
        for segment in segments:
            if split_morphs:
                resegs += segment.split('{MB}')
            else:
                resegs += [segment.replace('{MB}', '')]
        segments = resegs
        resegs = []
        for segment in segments:
            if split_words:
                resegs += segment.split('{WB}')
            else:
                resegs += [segment.replace('{WB}', '')]
        segments = resegs
        resegs = []
        for segment in segments:
            if split_new_words:
                resegs += segment.split('{wB}')
            else:
                resegs += [segment.replace('{wB}', '')]
        segments = resegs
        resegs = []
        for segment in segments:
            if split_derivs:
                resegs += segment.split('{DB}')
            else:
                resegs += [segment.replace('{DB}', '')]
        segments = resegs
        resegs = []
        for segment in segments:
            if split_nonwords:
                resegs += segment.split('{XB}')
            else:
                resegs += [segment.replace('{XB}', '')]
        return resegs

    def get_moses_factor_segments(self):
        '''Create moses factors from analyses.'''
        if self.rawtype != 'labelsegments':
            return None
        analysis = self.raw
        splat = re.split("[]{}[]", analysis)
        skiptag = None
        nextsep = '|'
        moses = ''
        allow_uppers = True
        for split in splat:
            if split == '':
                continue
            elif split in ['STUB', 'hyph?', 'XB']:
                allow_uppers = True
                continue
            elif split in ['SG', 'NOM', 'POS', 'ACTV', 'PRES']:
                # we actually skip 0 morphs...?
                allow_uppers = True
                continue
            elif split in ['DB', 'MB', 'WB', 'wB']:
                allow_uppers = True
                if skiptag:
                    moses += nextsep + skiptag
                    skiptag = None
                moses += ' '
                nextsep = '|'
            elif split in ['NOUN', 'VERB', 'ADJ', 'COMP', 'PROPN',
                           'SUPER', 'AUX', 'NUM', 'PRON', 'DET']:
                allow_uppers = True
                skiptag = split
            elif split in ['ADV', 'ADP', 'X', 'PUNCT', 'CCONJ',
                           'SCONJ', 'CCONJ|VERB', 'INTJ', 'SYM']:
                allow_uppers = True
                moses += nextsep + split
            elif split in ['PL', 'INS', 'INE', 'ELA',
                           'ILL', 'ADE', 'ABL', 'ALL', 'ACTV', 'PASV',
                           'IMPV', 'POTN', 'COND', 'SG1', 'SG2', 'SG3', 'PL1',
                           'PL2', 'PL3', 'PAST', 'INFA', 'PAR',
                           'POSSP3', 'POSSG1', 'POSSG2', 'POSPL1', 'POSPL2',
                           'GEN', 'PCPVA', 'INFE', 'PCPMA', 'PCPNUT', 'INFMA',
                           'PE4', 'ABE', 'ESS', 'CONNEG', 'ORD', 'TRA', 'COM',
                           'INFMAISILLA', 'PCPMATON',
                           'HAN', 'KO', 'PA', 'S', 'KAAN', 'KA', 'KIN',
                           'ACC']:
                allow_uppers = True
                if skiptag:
                    moses += nextsep + skiptag
                    skiptag = None
                    nextsep = '.'
                moses += nextsep + split
                nextsep = '.'
            elif split == 'TRUNC':
                allow_uppers = True
                # FIXME
                continue
            elif split.isupper():
                if not allow_uppers and not splat[0].startswith(split):
                    print("unhandlend upper string?", split, splat)
                    exit(1)
                else:
                    moses += split
                allow_uppers = False
            else:
                allow_uppers = False
                moses += split
        if skiptag:
            moses += nextsep + skiptag
        # tweaks and hacks
        if " i " in moses or " j " in moses:
            moses = re.sub(r" ([ij]) ([a-zä]*)\|PL.", r" \1|PL \2|", moses)
        # i ne|COM
        moses = re.sub(r"i ne\|COM", "i|PL ne|COM", moses)
        # |ABEko.KO
        moses = re.sub(r"\|([A-Z][A-Z][A-Z]?)ko\.KO", r"|\1 ko|KO", moses)
        moses = re.sub(r" ([a-zåäö]+) ", r" \1|NOUN ", moses)
        moses = re.sub(r"^([a-zåäö]+) ", r"\1|NOUN ", moses)
        moses = re.sub(r"([snrl])\|PCPNUTut", r"\1ut|PCPNUT", moses)
        moses = re.sub(r"([snrl])\|PCPNUTee", r"\1ee|PCPNUT", moses)
        moses = re.sub(
            r"([snrl])\|AUX\.PCPNUTee", r"\1ee|AUX.PASV.PCPNUT", moses)
        moses = re.sub(r"([snrl])\|PCPNUTe", r"\1e|PCPNUT", moses)
        moses = re.sub(
            r"([snrl])\|AUX\.PCPNUTe", r"\1e|AUX.PASV.PCPNUT", moses)
        # teh|VERB |PCPNUTdy llä|ADE
        moses = re.sub(r"\|PCPNUT([tdrsnl]?[uy])", r"\1|PCPNUT", moses)
        moses = re.sub(
            r"\|AUX\.PASV\.PCPNUT([tdrsnl]?[uy])", r"\1|AUX.PASV.PCPNUT",
            moses)
        moses = re.sub(r"m\|PCPMA([a-zaä]+)", r"m\1|PCPMA", moses)
        moses = re.sub(r"v\|PCPVA([a-zaä]+)", r"v\1|PCPVA", moses)
        moses = re.sub(r"v\|VERB\.PCPVA([aä])", r"v\1|VERB.PCPVA", moses)
        moses = re.sub(r"v\|AUX.PCPVA([aä])", r"v\1|AUX.PCPVA", moses)
        # |PCPMATONön
        moses = re.sub(r"\|PCPMATON([a-zåäö]+)", r"\1|PCPMATON", moses)
        # puhu|VERB ma|NOUN an|INFMA.ILL
        moses = re.sub(r"(m[aä])\|NOUN ([aä]n)\|INFMA.ILL",
                       r"\1|INFMA \2|ILL", moses)
        # esitet|VERB tä|NOUN mä n|PASV.INFMA.INS
        # esitet|VERB tä|NOUN mä n|PASV.INFMA.INS
        moses = re.sub(r"(t[aä])\|NOUN (m[aä]) n\|PASV\.INFMA\.INS",
                       r"\1|PASV \2|INFMA n|INS", moses)
        # annet|VERB ta|NOUN va t|PASV.PCPVA.PL
        moses = re.sub(r"(t[aä])\|NOUN (v[aä]) t\|PASV\.PCPVA\.PL",
                       r"\1|PASV \2|PCPVA t|PL", moses)
        # todistet|VERB ta|NOUN va sti|PASV.PCPVA
        moses = re.sub(r"(t[aä])\|NOUN (v[aä]) sti\|PASV\.PCPVA",
                       r"\1|PASV \2|PCPVA sti|STI", moses)
        moses = re.sub(
            r"([ei])\|NOUN (n|ssa|ssä)\|INFE.", r"\1|INFE \2|", moses)
        # herättäv|VERB.PCPVAä sti
        moses = re.sub(
            r"v\|VERB.PCPVA([aä]) sti", r"v\1|VERB.PCPVA sti|STI", moses)
        # tarkastel|VERB ta|NOUN e ssa|PASV.INFE.INE
        moses = re.sub(
            r"(t[aä])\|NOUN e (ssa|ssä)\|PASV.INFE.", r"\1|PASV e|INFE \2|",
            moses)
        # varot|VERB ta|NOUN e n|PASV.INFE.INS
        moses = re.sub(
            r"(t[aä])\|NOUN e n\|PASV.INFE.", r"\1|PASV e|INFE n|", moses)
        # tä|NOUN isi in|PASV.COND.PE4
        moses = re.sub(
            r"t([aä])\|NOUN isi in\|PASV.COND.PE4",
            r"t\1|PASV isi|COND in|PE4", moses)
        # moniselitteise|ADJ sti
        moses = re.sub(r"ADJ sti$", "ADJ sti|STI", moses)
        # hillitse|VERB vä|PCPVA sti
        moses = re.sub(r"PCPVA sti$", "PCPVA sti|STI", moses)
        # säästä|VERB väi|PCPVAs i|PL lle|ALL
        moses = re.sub(r"\|PCPVAs", r"s|PCPVA", moses)
        # valveutu|VERB nee|PCPNUT sti
        moses = re.sub(r"PCPNUT sti$", "PCPNUT sti|STI", moses)
        # ehdotta|VERB ma|PCPMA sti
        moses = re.sub(r"PCPMA sti$", "PCPMA sti|STI", moses)
        # yhdenmukaista|VERB minen
        moses = re.sub(r"\|VERB minen$", "|VERB minen|NOUN", moses)
        # mi|NOUN s
        moses = re.sub(r"mi\|NOUN s", "mis|NOUN", moses)
        # kunnalli|ADJ s-
        moses = re.sub(r"\|(ADJ|NOUN|PROPN) s-", r"s-|\1", moses)
        # tarvin|AUX ne|NOUN |POTN.CONNEG
        moses = re.sub(r"ne\|NOUN \|POTN.CONNEG", r"ne|POTN.CONNEG", moses)
        # kiittel|VERB y
        # rauhoittel|VERB u-
        moses = re.sub(r"VERB ([uy])$", r"VERB \1|NOUN", moses)
        moses = re.sub(r"VERB ([uy]-)$", r"VERB \1|NOUN", moses)
        # clusterfuckup:
        # soveltamis|NOUN|mis|NOUN|NOUN päivä määrä|NOUN n|GEN
        # siirtämis|NOUNs|NOUN tä|PAR
        moses = re.sub(r"mis\|NOUN\|mis\|NOUN\|NOUN ([a-zåäö]+)",
                       r"mis|NOUN \1|NOUN", moses)
        moses = re.sub(r"mis\|NOUNs\|NOUN", r"mis|NOUN", moses)
        # toimi|NOUN alo|NOUN i|NOUN ttain
        moses = re.sub(r"NOUN ttain$", "NOUN ttain|DERTTAIN", moses)
        # |X|NUM
        moses = re.sub(r"\|X\|NUM", r"X|NUM", moses)
        moses = re.sub(r"\|X-\|NUM", r"X-|NUM", moses)
        # e|CONJ|VERBmme
        moses = re.sub(r"CCONJ\|VERB", r"CCONJVERB ", moses)
        # elin|NOUN tarvike ala|NOUN
        # kehitys|NOUN yhteis|NOUN työ mis|NOUNnisteri|NOUN nä|ESS
        # ulko|NOUN maalai|NOUN s|NOUN viha mielis|ADJ tä|PAR
        # epä|NOUN tasa-arvo asia|NOUN
        moses = re.sub(
            r"([a-zéšäöå-]+)\|NOUN ([a-zšéäöå-]+) ([a-zšéäöå]+)\|NOUN",
            r"\1|NOUN \2|NOUN \3|NOUN", moses)
        moses = re.sub(
            r"([a-zéšäöå-]+)\|NOUN ([a-zšéäöå-]+) ([a-zšéäöå]+)\|ADJ",
            r"\1|NOUN \2|NOUN \3|ADJ", moses)
        # šakki lauda|NOUN
        # pöytä|NOUN rosé viine|NOUN i|PL stä|ELA
        # linja-auto liikentee|NOUN n|GEN
        moses = re.sub(r"^([A-Za-z*ÉéÄŠÖÅšäöå-]+) ([a-zäöå-]+)\|NOUN",
                       r"\1|NOUN \2|NOUN", moses)
        moses = re.sub(r"^([A-Za-z*ÉéÄŠÖÅšäöå-]+) ([a-zäöå-]+)\|PROPN",
                       r"\1|PROPN \2|PROPN", moses)
        moses = re.sub(r"^\|([A-Za-z*ÉéÄŠÖÅšäöå-]+).PROPN",
                       r"\1|PROPN", moses)
        #  R|NOUN ja|ADV |S.NOUN
        # |S-.NOUN kirjaime|NOUN lla|ADE
        moses = re.sub(r"\|S.NOUN", r"S|NOUN", moses)
        moses = re.sub(r"\|S-.NOUN", r"S-|NOUN", moses)
        # |PA.NOUN
        moses = re.sub(r"\|PA.NOUN", r"PA|NOUN", moses)
        # (|PUNCT |PL.NOUN )|PUNCT
        moses = re.sub(r"^\|PL.NOUN", r"PL|NOUN", moses)
        # |NOUN |NOUN Aerospacen|UNK
        moses = re.sub(r"^\|NOUN", r"", moses)
        moses = re.sub(r" \|NOUN", r"", moses)
        moses = re.sub(r"\|NOUN\|NOUN", r"|NOUN", moses)
        # |NOUNsa|INE
        moses = re.sub(r"\|NOUN([st][aä])\|", r"|NOUN \1|", moses)
        # Slovenia... n|GEN- haara
        moses = re.sub(r"n\|([A-Z.]+)-+", r"n|\1", moses)
        # ADP
        moses = re.sub(r"\|ADP([in])", r"\1|ADP", moses)
        moses = re.sub(r"\|SG3pi", r"pi|SG3", moses)
        #
        moses = re.sub(r"([uy])\|PCPNUTt", r"\1t|PCPNUT", moses)
        # |NOUNäiliö|NOUN
        moses = re.sub(r"\|NOUN([a-zäåö]+)\|NOUN", r"\1|NOUN", moses)
        # || special case :-/
        moses = re.sub(r"\|\|SYM", "@pipe;|SYM", moses)
        # finally
    ##    segleft = ''
    ##    segright = ''
    ##    if seglen == 0:
    ##        segleft = ''
    ##        segright = ''
    ##    elif seglen == 1:
    ##        segleft = options.segment_marker
    ##        segright = options.segment_marker
    ##    elif seglen % 2 == 0:
    ##        segleft = options.segment_marker[:int(seglen / 2)]
    ##        segright = options.segment_marker[int(seglen / 2):]
    ##    else:
    ##        segleft = options.segment_marker[:int((seglen - 1) / 2)]
    ##        segright = options.segment_marker[int((seglen - 1) / 2):]
    ##    moses = re.sub(r"\|", segleft + "|", moses)
    ##    moses = re.sub(r" ", " " + segright, moses)
    ##    last = moses.rfind(segleft + "|")
    ##    moses = moses[:last + len(segleft) - 1] + moses[last + len(segleft):]
        return moses.split()

    def get_ud_misc(self):
        '''Get random collection of analyses for token.

        Primarily used for UD MISC field but can be used for any extra data.
        '''
        miscs = []
        if self.manglers:
            miscs += ['|'.join(self.manglers)]
        if self.analsurf:
            miscs += ['AnalysisForm=' + self.analsurf]
        miscs += ['Weight=' + str(self.weight)]
        return miscs

    def printable_ud_misc(self):
        '''Formats UD misc like in UD data.'''
        miscs = self.get_ud_misc()
        if not miscs:
            return '_'
        return '|'.join(miscs)

    def printable_udepname(self):
        '''Format udep as string for CONLL-U.

        Returns:
            string of udep nam
        '''
        if self.udepname:
            return self.udepname
        else:
            return '_'

    def printable_udephead(self):
        '''Format udep head position for CONLL-U.

        Returns:
            string of non-ngative integer or _'''
        if self.udepname and self.udepname == 'root' and self.udeppos == 0:
            return '0'
        if self.udeppos:
            return str(self.udeppos)
        else:
            return '_'

    def printable_ud_feats(self, hacks=None):
        '''Formats UD feats from token data exactly as in fi-tdt data.

        When the correct analysis is in question the result should be equal
        to the UFEAT field of the connl-u data downloadable from UD web site,
        in string format.

        Returns:
            string of |-separated key=value pairs in correct order or _
        '''
        rvs = self.ufeats
        if not rvs:
            return '_'
        rv = ''
        for k in sorted(rvs, key=str.lower):
            rv += k + '=' + rvs[k] + '|'
        return rv.rstrip('|')

    def printable_unimorph(self):
        '''Formats FTB feats from token data like in FTB-2014 data.'''
        rvs = self.get_unimorph_feats()
        return ';'.join(rvs)

    def printable_ftb_feats(self):
        '''Formats FTB feats from token data like in FTB-2014 data.'''
        rvs = self.get_ftb_feats()
        return ' '.join(rvs)

    def get_xpos_ftb(self):
        '''Gets FTB-compatible part-of-speech from analysis.'''
        upos = self.get_upos()
        if upos in ['NOUN', 'PROPN']:
            return 'N'
        elif upos == 'ADJ':
            return 'A'
        elif upos in ['VERB', 'AUX']:
            # this is inverse that I intended but oh well...
            if 'PartForm' in self.ufeats:
                if self.ufeats['PartForm'] == 'Past':
                    return 'PrfPrc'
                elif self.ufeats['PartForm'] == 'Agt':
                    return 'AgPrc'
                elif self.ufeats['PartForm'] == 'Pres':
                    return 'PrsPrc'
                elif self.ufeats['PartForm'] == 'Neg':
                    return 'NegPrc'
                else:
                    return 'V'
            return 'V'
        elif upos == 'CCONJ':
            return 'CC'
        elif upos == 'SCONJ':
            return 'CS'
        elif upos == 'ADP':
            return 'Adp'
        elif upos == 'ADV':
            return 'Adv'
        elif upos == 'PRON':
            return 'Pron'
        elif upos in ['PUNCT', 'SYM']:
            return 'Punct'
        elif upos == 'INTJ':
            return 'Interj'
        elif upos == 'NUM':
            return 'Num'
        elif upos == 'DET':
            return 'Pron'
        elif upos == 'X':
            return 'Forgn'
        else:
            return 'Unkwn'
        return upos

    def get_xpos_tdt(self):
        '''Get TDT-compatible part-of-speech from analysed token.'''
        upos = self.get_upos()
        if upos in ['NOUN', 'PROPN']:
            return 'N'
        elif upos == 'ADJ':
            return 'A'
        elif upos in ['VERB', 'AUX']:
            return 'V'
        elif upos in ['CCONJ', 'SCONJ']:
            return 'C'
        elif upos == 'ADP':
            return 'Adp'
        elif upos == 'ADV':
            return 'Adv'
        elif upos == 'PRON':
            return 'Pron'
        elif upos == 'PUNCT':
            return 'Punct'
        elif upos == 'SYM':
            return 'Symb'
        elif upos == 'INTJ':
            return 'Interj'
        elif upos == 'NUM':
            return 'Num'
        else:
            return 'X'

    def printable_vislcg(self):
        '''Create VISL-CG 3 output from the token.'''
        mrds = self.get_vislcg_feats()
        lemmas = self.get_lemmas()
        return '\t"' + '#'.join(lemmas) + '" ' + ' '.join(mrds)

    def is_oov(self):
        '''Figures out if this analysis was guessed for an OOV.'''
        if self.manglers:
            for mangler in self.manglers:
                if 'GUESS' in mangler.upper():
                    return True
        if self.weight == float("inf"):
            return True
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test sort | uniq -c | sort -nr'd gold analysis data for faithfulness
"""


from argparse import ArgumentParser, FileType
from sys import stderr, stdout, stdin
from time import perf_counter, process_time

from omorfi import Omorfi, Token


def main():
    """Command-line interface for omorfi's sort | uniq -c tester."""
    a = ArgumentParser()
    a.add_argument('-a', '--analyser', metavar='FSAFILE', required=True,
                   help="load analyser from FSAFILE")
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-o', '--output', metavar="OUTFILE",
                   type=FileType('w'),
                   dest="outfile", help="log outputs to OUTFILE")
    a.add_argument('-X', '--statistics', metavar="STATFILE",
                   type=FileType('w'),
                   dest="statfile", help="statistics")
    a.add_argument('-v', '--verbose', action="store_true", default=False,
                   help="Print verbosely while processing")
    a.add_argument('-C', '--no-casing', action="store_true", default=False,
                   help="Do not try to recase input and output when matching")
    a.add_argument('-f', '--format', metavar="FORMAT",
                   help="use FORMAT formatter to compare analyses",
                   choices=["coverage", "ftb3.1"], default="coverage")
    a.add_argument('-c', '--count', metavar="FREQ", default=0,
                   help="test only word-forms with frequency higher than FREQ")
    a.add_argument('-t', '--threshold', metavar="THOLD", default=99,
                   help="if coverage is less than THOLD exit with error")
    a.add_argument('-O', '--no-workaround', default=False, action="store_true",
                   help="do not skip suspicious strings (that may break)")
    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    try:
        if options.analyser:
            if options.verbose:
                print("reading analyser from", options.analyser)
            omorfi.load_analyser(options.analyser)
        if not options.infile:
            options.infile = stdin
            print("reading from <stdin>")
        if not options.statfile:
            options.statfile = stdout
        if not options.outfile:
            options.outfile = stdout
    except IOError:
        print("Could not process file", options.analyser, file=stderr)
        exit(2)
    # basic statistics
    covered = 0
    full_matches = 0
    lemma_matches = 0
    anal_matches = 0
    only_permuted = 0
    only_rehashed = 0
    no_matches = 0
    no_results = 0
    lines = 0
    # types
    types_covered = 0
    types_no_results = 0
    types = 0
    # for make check target
    threshold = options.threshold
    realstart = perf_counter()
    cpustart = process_time()
    for line in options.infile:
        fields = line.strip().replace(' ', '\t', 1).split('\t')
        if len(fields) < 2:
            print("ERROR: Skipping line", fields, file=stderr)
            continue
        freq = int(fields[0])
        if freq < int(options.count):
            break
        surf = fields[1]
        if len(surf) > 120 and not options.no_workaround:
            print("WARN: Skipping line", fields, file=stderr)
            continue
        lemma = surf
        analysis = surf
        if options.format != 'coverage':
            lemma = fields[2]
            analysis = fields[3]
        lines += freq
        types += 1
        if options.verbose:
            print(lines, '(', freq, ') ...', end='\r')
        token = Token(surf)
        # pos 1 triggers acceptable detitlecasing
        token.pos = 1
        omorfi.analyse(token)
        if token.is_oov():
            omorfi.guess(token)
        if not token.is_oov():
            covered += freq
            types_covered += 1
        else:
            no_results += freq
            types_no_results += 1
            print(freq, "OOV", surf, sep='\t', file=options.outfile)
        found_anals = False
        found_lemma = False
        rehashed = True
        permuted = True
        for anal in token.analyses:
            if options.format == 'ftb3.1':
                anal_ftb3 = ' '.join(anal.get_ftb_feats())
                lemma_ftb3 = '#'.join(anal.get_lemmas())
                # hacks ftb3:
                analysis = analysis.replace(" >>>", "")
                if analysis == anal_ftb3:
                    found_anals = True
                    permuted = False
                elif set(anal_ftb3.split()) == set(analysis.split()):
                    found_anals = True
                    print(freq, "PERMUTAHIT", analysis, anal_ftb3, sep='\t',
                          file=options.outfile)
                else:
                    print(freq, "ANALMISS", analysis, anal_ftb3, sep='\t',
                          file=options.outfile)
                if lemma == lemma_ftb3:
                    found_lemma = True
                    rehashed = False
                elif lemma.replace('#', '') == lemma_ftb3.replace('#', ''):
                    found_lemma = True
                    print(freq, "LEMMARECOMP", lemma, lemma_ftb3, sep='\t',
                          file=options.outfile)
                else:
                    print(freq, "LEMMAMISS", lemma, lemma_ftb3, sep='\t',
                          file=options.outfile)
        if options.format != 'coverage':
            if not found_anals and not found_lemma:
                no_matches += freq
                print(freq, "NOHITS!", surf, sep='\t', file=options.outfile)
            elif found_anals and found_lemma:
                full_matches += freq
            elif not found_anals:
                anal_matches += freq
                print(freq, "LEMMANOANAL", surf, sep='\t',
                      file=options.outfile)
            elif not found_lemma:
                lemma_matches += freq
                print(freq, "ANALNOLEMMA", surf, sep='\t',
                      file=options.outfile)
            else:
                print("Logical error, kill everyone")
                exit(13)
            if rehashed:
                only_rehashed += freq
            if permuted:
                only_permuted += freq
    realend = perf_counter()
    cpuend = process_time()
    print("CPU time:", cpuend - cpustart, "real time:", realend - realstart)
    print("Lines", "Covered", "OOV", sep="\t", file=options.statfile)
    print(lines, covered, lines - covered, sep="\t", file=options.statfile)
    print(lines / lines * 100 if lines != 0 else 0,
          covered / lines * 100 if lines != 0 else 0,
          (lines - covered) / lines * 100 if lines != 0 else 0,
          sep="\t", file=options.statfile)
    print("Types", "Covered", "OOV", sep="\t", file=options.statfile)
    print(types, types_covered, types - types_covered, sep="\t",
          file=options.statfile)
    print(types / types * 100 if types != 0 else 0,
          types_covered / types * 100 if types != 0 else 0,
          (types - types_covered) / types * 100 if types != 0 else 0,
          sep="\t", file=options.statfile)
    if options.format == 'ftb3.1':
        print("Lines", "Matches", "Lemma", "Anals", "Mismatch",
              "No results", sep="\t", file=options.statfile)
        print(lines, full_matches, lemma_matches, anal_matches, no_matches,
              no_results,
              sep="\t", file=options.statfile)
        print(lines / lines * 100 if lines != 0 else 0,
              full_matches / lines * 100 if lines != 0 else 0,
              lemma_matches / lines * 100 if lines != 0 else 0,
              anal_matches / lines * 100 if lines != 0 else 0,
              no_matches / lines * 100 if lines != 0 else 0,
              no_results / lines * 100 if lines != 0 else 0,
              sep="\t", file=options.statfile)
        print("Of which", "Tag permuations", "Lemma rehashing", sep='\t',
              file=options.statfile)
        print(lines / lines * 100 if lines != 0 else 0,
              only_permuted / lines * 100 if lines != 0 else 0,
              only_rehashed / lines * 100 if lines != 0 else 0, sep='\t',
              file=options.statfile)
    if lines == 0:
        print("Needs more than 0 lines to determine something",
              file=stderr)
        exit(2)
    elif options.format == 'ftb3.1' and \
            (full_matches / lines * 100 <= int(options.threshold)):
        print("needs to have", threshold, "% matches to pass regress test\n",
              "please examine", options.outfile.name, "for regressions",
              file=stderr)
        exit(1)
    elif options.format == 'coverage' and \
            (covered / lines * 100 <= int(options.threshold)):
        print("needs to have", threshold, "% coverage to pass regress test\n",
              "please examine", options.outfile.name, "for regressions",
              file=stderr)
        exit(1)
    else:
        exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Command-line interface to create moses factored analyses with omorfi."""

from argparse import ArgumentParser
from sys import stdin, stdout, stderr

from omorfi.omorfi import Omorfi


def main():
    """Preprocess text for moses factored modeling."""
    a = ArgumentParser()
    a.add_argument('-a', '--analyser', metavar='AFILE',
                   help="load analyser model from AFILE", required=True)
    a.add_argument('-s', '--segmenter', metavar='SFILE',
                   help="load segmenter model from SFILE", required=True)
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-v', '--verbose', action='store_true',
                   help="print verbosely while processing")
    a.add_argument('-o', '--output', metavar="OUTFILE",
                   help="print factors into OUTFILE")
    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    if options.analyser:
        if options.verbose:
            print("Reading analyser model", options.analyser)
        omorfi.load_analyser(options.analyser)
    else:
        print("at least analyser file is needed", file=stderr)
        exit(1)
    if options.segmenter:
        if options.verbose:
            print("Reading segmenter model", options.segmenter)
        omorfi.load_segmenter(options.segmenter)
    else:
        print("at least segmenter file is needed", file=stderr)
        exit(1)
    if options.infile:
        infile = options.infile
    else:
        infile = stdin
    if options.output:
        outfile = open(options.output, 'w')
    else:
        outfile = stdout
    if options.verbose:
        print("reading from", infile.name)
    if options.verbose:
        print("writign to", outfile.name)
    linen = 0
    for line in infile:
        line = line.strip()
        linen += 1
        if options.verbose and linen % 10000 == 0:
            print(linen, '...')
        if not line or line == '':
            continue
        tokens = omorfi.tokenise_sentence(line)
        for token in tokens:
            if not token.surf:
                continue
            anals = omorfi.analyse(token)
            pos = "X"
            mrds = ["?"]
            lemmas = [token.surf]
            if anals:
                anal = token.get_best()
                pos = anal.get_upos()
                mrds = anal.get_ufeats()
                lemmas = anal.get_lemmas()
            segments = omorfi.segment(token)
            morphs = "0"
            if segments:
                segment = token.get_best_segments()
                if segment:
                    parts = segment.get_segments()
                    morphs = ".".join(parts)
                else:
                    morphs = token.surf
            print(token.surf, '+'.join(lemmas), pos, '.'.join(mrds),
                  morphs, sep='|', end=' ', file=outfile)
        print(file=outfile)
    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Command-line interface for omorfi analysis in VISL CG-3 formats."""

# string munging
from argparse import ArgumentParser, FileType
# CLI stuff
from sys import stdin, stdout, stderr
# statistics
from time import perf_counter, process_time

# omorfi
from omorfi import Omorfi
from omorfi.formats.fileformats import next_conllu, next_vislcg, next_plaintext
from omorfi.disamparsulate import Disamparsulator


def main():
    """Invoke a simple CLI analyser."""
    a = ArgumentParser()
    a.add_argument('-a', '--analyser', metavar='AFILE',
                   help="load analyser model from AFILE")
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-v', '--verbose', action='store_true',
                   help="print verbosely while processing")
    a.add_argument('-o', '--output', metavar="OUTFILE", dest="outfile",
                   help="print output into OUTFILE", type=FileType('w'))
    a.add_argument('-F', '--format', metavar="INFORMAT", default='text',
                   help="read input using INFORMAT tokenisation",
                   choices=['text', 'vislcg', 'conllu', 'sentences'])
    a.add_argument('-x', '--statistics', metavar="STATFILE", dest="statfile",
                   help="print statistics to STATFILE", type=FileType('w'))
    a.add_argument('--not-rules', metavar="RULEFILE", type=open,
                   help="read non-rules from RULEFILE")
    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    if options.analyser:
        if options.verbose:
            print("reading analyser model", options.analyser)
        omorfi.load_analyser(options.analyser)
    else:
        print("analyser is required to vislcg", file=stderr)
        exit(4)
    disamparsulator = None
    if options.not_rules:
        if options.verbose:
            print("Reading rulestuff", options.not_rules.name)
        disamparsulator = Disamparsulator()
        disamparsulator.frobblesnizz(options.not_rules)
    if not options.infile:
        options.infile = stdin
    if options.verbose:
        print("analysing", options.infile.name)
    if not options.outfile:
        options.outfile = stdout
    if options.verbose:
        print("writing to", options.outfile.name)
    if not options.statfile:
        if options.outfile == stdout:
            options.statfile = stdout
        else:
            options.statfile = stderr
    # statistics
    realstart = perf_counter()
    cpustart = process_time()
    tokencount = 0
    unknowns = 0
    eoffed = False
    while not eoffed:
        if options.format == 'vislcg':
            tokens = next_vislcg(options.infile)
        elif options.format == 'text':
            tokens = next_plaintext(options.infile)
        elif options.format == 'conllu':
            tokens = next_conllu(options.infile)
        else:
            print("input format missing implementation", options.format,
                  file=stderr)
            exit(2)
        if not tokens:
            break
        for token in tokens:
            if token.surf:
                tokencount += 1
                omorfi.analyse(token)
                if token.is_oov():
                    unknowns += 1
                    omorfi.guess(token)
            elif token.error or token.nontoken:
                pass
            else:
                print("Unrecognised", token, file=stderr)
                exit(2)
        if disamparsulator:
            disamparsulator.linguisticate(tokens)
        for token in tokens:
            if token.nontoken and token.nontoken == "eof":
                eoffed = True
                break
            print(token.printable_vislcg(), file=options.outfile)
    cpuend = process_time()
    realend = perf_counter()
    print("# Tokens:", tokencount, "\n# Unknown:", unknowns,
          unknowns / tokencount * 100 if tokencount > 0 else 0, "%",
          file=options.statfile)
    print("# CPU time:", cpuend - cpustart,
          "\n# Real time:", realend - realstart,
          file=options.statfile)
    print("# Tokens per timeunit:", tokencount / (realend - realstart),
          file=options.statfile)
    exit(0)


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="omorfi", # Replace with your own username
    version="0.0.2",
    author="Flammie A Pirinen",
    author_email="flammie@iki.fi",
    description="Open morphology for Finnish, python bindings",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/flammie/omorfi",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    keywords="finnish nlp morphology",
    python_requires='>=3.5',
    scripts=['omorfi-download.py',
             'omorfi-conllu.py',
             'omorfi-convert.py',
             'omorfi-disamparsulate.py',
             'omorfi-download.py',
             'omorfi-factorise.py',
             'omorfi-freq-evals.py',
             'omorfi-ftb3.py',
             'omorfi-segment.py',
             'omorfi-sigmorphons.py',
             'omorfi-tokenise.py',
             'omorfi-unimorph.py',
             'omorfi-vislcg.py',
             'omorfi-wikitable.py'],
    install_requires=['hfst'],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions to guess plurale tantum features from partial lexical data."""

# Author: Omorfi contributors <omorfi-devel@groups.google.com> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .error_logging import fail_guess_because
from .omorfi.string_manglers import replace_rightmost, replace_rightmosts


def plurale_tantum_get_singular_stem(wordmap):
    '''Guess inflectional singulars for words whose dictionary form is plural.
    '''
    if not wordmap['kotus_tn']:
        return wordmap
    tn = int(wordmap['kotus_tn'])
    if not wordmap['plurale_tantum'] == 'obligatory':
        return wordmap
    if wordmap['kotus_av'] and tn not in [27, 28, 34]:
        if tn in [7, 23, 24, 25, 26, 29, 30]:
            wordmap['stub'] = replace_rightmost(wordmap['stub'], 'e', 'i')
        # ^^ generic extra tweak
        if (tn in range(1, 7) or tn in range(8, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'A':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'k', 'kk')
        elif (tn in range(1, 7) or tn in range(8, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'B':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'p', 'pp')
        elif (tn in range(1, 7) or tn in range(8, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'C':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 't', 'tt')
        elif tn == 1 and wordmap['kotus_av'] == 'D':
            wordmap['stub'] = replace_rightmosts(
                wordmap['stub'], ['ut', 'yt', 'öt', 'ot'], ['ku', 'ky', 'kö', 'ko'])
        elif tn in [5, 7] and wordmap['kotus_av'] == 'D':
            wordmap['stub'] = replace_rightmost(wordmap['stub'], 'it', 'ki')
        elif tn in [9, 10] and wordmap['kotus_av'] == 'D':
            wordmap['stub'] = replace_rightmosts(
                wordmap['stub'], ['at', 'ät'], ['ka', 'kä'])
        elif tn == 1007:
            wordmap['stub'] = replace_rightmost(wordmap['stub'], 'jet', 'i')
        elif tn in [1009, 1010] and wordmap['kotus_av'] == 'D':
            wordmap['stub'] = replace_rightmost(wordmap['stub'], 'jat', 'ika')
        elif tn == 32 and wordmap['kotus_av'] == 'C':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 'et', ''), 'tt', 't')
        elif tn == 32 and wordmap['kotus_av'] == 'D':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 'et', ''), 'k', '')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'E':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'v', 'p')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'F':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'd', 't')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'G':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'g', 'k')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'H':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'mm', 'mp')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'I':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'll', 'lt')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'J':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'nn', 'nt')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'K':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'rr', 'rt')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'L':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'j', 'k')
        elif (tn in range(1, 16) or tn in range(17, 22)) and wordmap['kotus_av'] == 'M':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'v', 'k')
        elif (tn in [23, 24, 25, 26, 29, 30]) and wordmap['kotus_av'] == 'G':
            wordmap['stub'] = replace_rightmost(
                replace_rightmost(wordmap['stub'], 't', ''), 'g', 'k')
        elif tn == 16 and wordmap['kotus_av'] == 'H':
            wordmap['stub'] = wordmap['stub'][:-3] + 'pi'
        elif tn == 28 and wordmap['kotus_av'] == 'J':
            wordmap['stub'] = replace_rightmost(wordmap['stub'], 'nnet', 'nsi')
        elif tn == 33 and wordmap['kotus_av'] in ['A', 'B', 'C', 'D']:
            wordmap['stub'] = wordmap['stub'][:-5] + 'in'
        elif tn == 33 and wordmap['kotus_av'] == 'E':
            wordmap['stub'] = wordmap['stub'][:-5] + 'vin'
        elif tn == 33 and wordmap['kotus_av'] == 'F':
            wordmap['stub'] = wordmap['stub'][:-5] + 'din'
        elif tn == 33 and wordmap['kotus_av'] == 'G':
            wordmap['stub'] = wordmap['stub'][:-5] + 'gin'
        elif tn == 33 and wordmap['kotus_av'] == 'H':
            wordmap['stub'] = wordmap['stub'][:-5] + 'min'
        elif tn == 33 and wordmap['kotus_av'] == 'I':
            wordmap['stub'] = wordmap['stub'][:-5] + 'lin'
        elif tn == 33 and wordmap['kotus_av'] == 'J':
            wordmap['stub'] = wordmap['stub'][:-5] + 'nin'
        elif tn == 33 and wordmap['kotus_av'] == 'K':
            wordmap['stub'] = wordmap['stub'][:-5] + 'rin'
        elif tn == 41 and wordmap['kotus_av'] in ['A', 'B', 'C', 'D']:
            wordmap['stub'] = wordmap['stub'][:-4] + 'as'
        elif tn == 41 and wordmap['kotus_av'] == 'E':
            wordmap['stub'] = wordmap['stub'][:-4] + 'vas'
        elif tn == 41 and wordmap['kotus_av'] == 'F':
            wordmap['stub'] = wordmap['stub'][:-4] + 'das'
        elif tn == 41 and wordmap['kotus_av'] == 'G':
            wordmap['stub'] = wordmap['stub'][:-4] + 'gas'
        elif tn == 41 and wordmap['kotus_av'] == 'H':
            wordmap['stub'] = wordmap['stub'][:-4] + 'mas'
        elif tn == 41 and wordmap['kotus_av'] == 'I':
            wordmap['stub'] = wordmap['stub'][:-4] + 'las'
        elif tn == 41 and wordmap['kotus_av'] == 'J':
            wordmap['stub'] = wordmap['stub'][:-4] + 'nas'
        elif tn == 41 and wordmap['kotus_av'] == 'K':
            wordmap['stub'] = wordmap['stub'][:-4] + 'ras'
        elif tn == 48 and wordmap['kotus_av'] in ['A', 'B', 'C', 'D']:
            wordmap['stub'] = wordmap['stub'][:-4] + 'e'
        elif tn == 48 and wordmap['kotus_av'] == 'E':
            wordmap['stub'] = wordmap['stub'][:-4] + 've'
        elif tn == 48 and wordmap['kotus_av'] == 'F':
            wordmap['stub'] = wordmap['stub'][:-4] + 'de'
        elif tn == 48 and wordmap['kotus_av'] == 'G':
            wordmap['stub'] = wordmap['stub'][:-4] + 'ge'
        elif tn == 48 and wordmap['kotus_av'] == 'H':
            wordmap['stub'] = wordmap['stub'][:-4] + 'me'
        elif tn == 48 and wordmap['kotus_av'] == 'I':
            wordmap['stub'] = wordmap['stub'][:-4] + 'le'
        elif tn == 48 and wordmap['kotus_av'] == 'J':
            wordmap['stub'] = wordmap['stub'][:-4] + 'ne'
        elif tn == 48 and wordmap['kotus_av'] == 'K':
            wordmap['stub'] = wordmap['stub'][:-4] + 're'
        elif tn == 48 and wordmap['kotus_av'] == 'L':
            wordmap['stub'] = wordmap['stub'][:-4] + 'je'
        else:
            fail_guess_because(wordmap, ["av", "!27,28,34"], ["1-48"])
            exit(1)
    else:
        if tn in range(1, 7) or tn in range(8, 16) or tn in range(17, 22):
            wordmap['stub'] = replace_rightmost(wordmap['stub'], 't', '')
        elif tn in [7, 23, 24, 25, 26, 29, 30]:
            wordmap['stub'] = replace_rightmost(wordmap['stub'], 'et', 'i')
        elif tn == 16:
            wordmap['stub'] = wordmap['stub'][:-3] + 'pi'
        elif tn == 16:
            wordmap['stub'] = wordmap['stub'][:-2]
        elif tn in [27, 28]:
            wordmap['stub'] = wordmap['stub'][:-3] + 'si'
            wordmap['kotus_av'] = False
        elif tn == 31:
            wordmap['stub'] = wordmap['stub'][:-4] + 'ksi'
        elif tn == 32:
            wordmap['stub'] = wordmap['stub'][:-2]
        elif tn == 33:
            wordmap['stub'] = wordmap['stub'][:-3] + 'n'
        elif tn == 34:
            wordmap['stub'] = wordmap['stub'][:-5] + '{oö}n'
        elif tn == 35:
            wordmap['stub'] = wordmap['stub'][:-5] + 'min'
        elif tn == 36 or tn == 37:
            wordmap['stub'] = wordmap['stub'][:-4] + 'n'
        elif tn == 38:
            wordmap['stub'] = wordmap['stub'][:-3] + 'nen'
        elif tn == 39:
            wordmap['stub'] = wordmap['stub'][:-4] + 's'
        elif tn == 40:
            wordmap['stub'] = wordmap['stub'][:-3] + 's'
        elif tn == 41:
            wordmap['stub'] = wordmap['stub'][:-2] + 's'
        elif tn == 42:
            wordmap['stub'] = wordmap['stub'][:-3] + 's'
        elif tn == 43:
            wordmap['stub'] = wordmap['stub'][:-2] + 't'
        elif tn == 47:
            wordmap['stub'] = wordmap['stub'][:-3] + '{uy}t'
        elif tn == 48:
            wordmap['stub'] = wordmap['stub'][:-3] + 'e'
        elif tn == 49:
            wordmap['stub'] = wordmap['stub'][:-3]
        else:
            fail_guess_because(wordmap, ["!av"], ["1-49"])
            exit(1)

    if wordmap['harmony'] == 'front':
        wordmap['stub'] = wordmap['stub'].replace('{uy}', 'y')
        wordmap['stub'] = wordmap['stub'].replace('{oö}', 'ö')
    elif wordmap['harmony'] == 'back':
        wordmap['stub'] = wordmap['stub'].replace('{uy}', 'u')
        wordmap['stub'] = wordmap['stub'].replace('{oö}', 'o')

    wordmap['gradestem'] = wordmap['stub']
    return wordmap



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test omorfi - unimorph compatibility
"""


from argparse import ArgumentParser, FileType
from sys import stderr, stdout, stdin
from time import perf_counter, process_time

from omorfi import Omorfi

u2o = {'pos=N': '[UPOS=NOUN]',
       'pos=V': '[UPOS=VERB]',
       'pos=ADJ': '[UPOS=ADJ]',
       'case=NOM': '[CASE=NOM]',
       'case=ON+ESS': '[CASE=ADE]',
       'case=ON+ABL': '[CASE=ABL]',
       'case=PRIV': '[CASE=ABE]',
       'case=IN+ESS': '[CASE=INE]',
       'case=IN+LAT': '[CASE=ILL]',
       'case=IN+ABL': '[CASE=ELA]',
       'case=ACC': '[CASE=ACC]',
       'case=PRT': '[CASE=PAR]',
       'case=COM': '[CASE=COM]',
       'case=GEN': '[CASE=GEN]',
       'case=INS': '[CASE=INS]',
       'case=FRML': '[CASE=ESS]',
       'case=TRANS': '[CASE=TRA]',
       'case=ON+ALL': '[CASE=ALL]',
       'polar=POS': '',
       'mood=IMP': '[MOOD=IMPV]',
       'mood=IND': '[MOOD=INDV]',
       'mood=COND': '[MOOD=COND]',
       'mood=POT': '[MOOD=POTN]',
       'mood=PURP': '[INF=A][CASE=TRA][POSS=3]',
       'aspect=PROSP': '[DRV=MAISILLA][POSS=3]',
       'finite=NFIN': '[INF=A][NUM=SG][CASE=LAT]',
       'tense=PRS': '[TENSE=PRESENT]',
       'tense=PST': '[TENSE=PAST]',
       'voice=ACT': '[VOICE=ACT]',
       'voice=PASS': '[VOICE=PSS]',
       'comp=SPRL': '[DRV=IN²][CMP=SUP]',
       'comp=CMPR': '[DRV=MPI][CMP=CMP]',
       'per=1': '[PERS=__1]',
       'per=2': '[PERS=__2]',
       'per=3': '[PERS=__3]',
       'num=SG': '[NUM=SG]',
       'num=PL': '[NUM=PL]'
       }


def unimorph2omor(unimorphstring):
    unimorphs = unimorphstring.split(',')
    omors = ''
    for unimorph in unimorphs:
        if 'pos' not in unimorph:
            continue
        if unimorph in u2o:
            omors += u2o[unimorph]
        else:
            print("missing", unimorph)
            exit(1)
    for unimorph in unimorphs:
        if 'comp' not in unimorph and 'voice' not in unimorph:
            continue
        if unimorph in u2o:
            omors += u2o[unimorph]
        else:
            print("missing", unimorph)
            exit(1)
    for unimorph in unimorphs:
        if 'mood' not in unimorph and 'aspect' not in unimorph and 'finite' \
                not in unimorph:
            continue
        if unimorph in u2o:
            omors += u2o[unimorph]
        else:
            print("missing", unimorph)
            exit(1)
    for unimorph in unimorphs:
        if 'tense' not in unimorph:
            continue
        if unimorph in u2o:
            omors += u2o[unimorph]
        else:
            print("missing", unimorph)
            exit(1)
    for unimorph in unimorphs:
        if 'per' not in unimorph:
            continue
        if unimorph in u2o:
            omors += u2o[unimorph]
        else:
            print("missing", unimorph)
            exit(1)
    for unimorph in unimorphs:
        if 'num' not in unimorph:
            continue
        if unimorph in u2o:
            omors += u2o[unimorph]
        else:
            print("missing", unimorph)
            exit(1)
    for unimorph in unimorphs:
        if 'case' not in unimorph:
            continue
        if unimorph in u2o:
            omors += u2o[unimorph]
        else:
            print("missing", unimorph)
            exit(1)
    omors = omors.replace("UPOS=ADJ][NUM", "UPOS=ADJ][CMP=POS][NUM")
    omors = omors.replace("UPOS=VERB][MOOD", "UPOS=VERB][VOICE=ACT][MOOD")
    omors = omors.replace("PERS=__1][NUM=SG]", "PERS=SG1]")
    omors = omors.replace("PERS=__2][NUM=SG]", "PERS=SG2]")
    omors = omors.replace("PERS=__3][NUM=SG]", "PERS=SG3]")
    omors = omors.replace("PERS=__1][NUM=PL]", "PERS=PL1]")
    omors = omors.replace("PERS=__2][NUM=PL]", "PERS=PL2]")
    omors = omors.replace("PERS=__3][NUM=PL]", "PERS=PL3]")
    omors = omors.replace("COND][TENSE=PAST]", "COND]")
    omors = omors.replace("IMPV][TENSE=PAST]", "IMPV]")
    omors = omors.replace("POTN][TENSE=PAST]", "POTN]")
    omors = omors.replace("IMPV][TENSE=PRESENT]", "IMPV]")
    omors = omors.replace("POTN][TENSE=PRESENT]", "POTN]")
    omors = omors.replace("COND][TENSE=PRESENT]", "COND]")
    omors = omors.replace("NUM=PL][CASE=COM]", "CASE=COM]")
    if 'VOICE=PSS' in omors:
        omors += '[PERS=PE4]'
    if 'CMP=SUP' in omors or 'CMP=CMP' in omors:
        omors += '[NUM=SG][CASE=NOM]'
    if 'CASE=ACC' in omors:
        if 'NUM=SG' in omors:
            omors = omors.replace("CASE=ACC", "CASE=GEN")
        elif "NUM=PL" in omors:
            omors = omors.replace("CASE=ACC", "CASE=NOM")
        else:
            print("ARGEG¤EFREAF¤EFFWA")
            exit(2)
    if 'CASE=COM' in omors and 'UPOS=NOUN' in omors:
        # nonsense but argh
        omors += '[POSS=3]'
    return omors


def main():
    """Command-line interface for omorfi's sort | uniq -c tester."""
    a = ArgumentParser()
    a.add_argument('-a', '--analyser', metavar='FSAFILE', required=True,
                   help="load analyser from FSAFILE")
    a.add_argument('-g', '--generator', metavar='FSAFILE', required=True,
                   help="load analyser from FSAFILE")
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-o', '--output', metavar="OUTFILE",
                   type=FileType('w'),
                   dest="outfile", help="log outputs to OUTFILE")
    a.add_argument('-X', '--statistics', metavar="STATFILE",
                   type=FileType('w'),
                   dest="statfile", help="statistics")
    a.add_argument('-v', '--verbose', action="store_true", default=False,
                   help="Print verbosely while processing")
    a.add_argument('-C', '--no-casing', action="store_true", default=False,
                   help="Do not try to recase input and output when matching")
    a.add_argument('-t', '--threshold', metavar="THOLD", default=99,
                   help="if coverage is less than THOLD exit with error")
    a.add_argument('-F', '--format', metavar="FMT", required=True,
                   help="which SIGMORHON shared task format is used")

    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    try:
        if options.analyser:
            if options.verbose:
                print("reading analyser from", options.analyser)
            omorfi.load_analyser(options.analyser)
        if options.generator:
            if options.verbose:
                print("reading generator from", options.generator)
            omorfi.load_generator(options.generator)
        if not options.infile:
            options.infile = stdin
            print("reading from <stdin>")
        if not options.statfile:
            options.statfile = stdout
        if not options.outfile:
            options.outfile = stdout
    except IOError:
        print("Could not process file", options.analyser, file=stderr)
        exit(2)
    # basic statistics
    correct = 0
    incorrect = 0
    oov = 0
    lines = 0
    # for make check target
    realstart = perf_counter()
    cpustart = process_time()
    for line in options.infile:
        fields = line.strip().split('\t')
        if len(fields) < 3:
            print("ERROR: Skipping line", fields, file=stderr)
            continue
        omors = None
        lemma = None
        print("<<<", fields)
        if options.format == '1':
            lemma = fields[0]
            omors = unimorph2omor(fields[1])
        elif options.format == '2':
            srcomors = unimorph2omor(fields[0])
            srchyps = omorfi.analyse(fields[1])
            for srchyp in srchyps:
                if srcomors in srchyp.raw and len(srchyp.get_lemmas()) == 1:
                    lemma = srchyp.get_lemmas()[0]
            if not lemma:
                lemma = ''.join(srchyps[0].get_lemmas())
            omors = unimorph2omor(fields[2])
        elif options.format == '3':
            srchyps = omorfi.analyse(fields[0])
            for srchyp in srchyps:
                if len(srchyp.get_lemmas()) == 1:
                    lemma = srchyp.get_lemmas()[0]
            if not lemma:
                lemma = ''.join(srchyps[0].get_lemmas())
            omors = unimorph2omor(fields[1])
        else:
            print("format fail", options.format)
            exit(1)
        genomor = '[WORD_ID=' + lemma + ']' + omors
        print(">>> ", genomor)
        generations = omorfi.generate(genomor)
        if not generations or '[' in generations:
            oov += 1
            genat1 = lemma
            print("OOV", genat1)
        else:
            genat1 = generations.split('/')[0]
            print("@1 ", genat1)
        if options.format == '1':
            if genat1 == fields[2]:
                correct += 1
            else:
                print("MIS", genat1, "!=", fields[2])
                incorrect += 1
        elif options.format == '2':
            if genat1 == fields[3]:
                correct += 1
            else:
                print("MIS", genat1, "!=", fields[2])
                incorrect += 1
        elif options.format == '3':
            if genat1 == fields[2]:
                correct += 1
            else:
                print("MIS", genat1, "!=", fields[2])
                incorrect += 1
        lines += 1
        if options.verbose and lines % 1000 == 0:
            print(lines, '...')
    realend = perf_counter()
    cpuend = process_time()
    print("CPU time:", cpuend - cpustart, "real time:", realend - realstart)
    if lines == 0:
        print("Needs more than 0 lines to determine something",
              file=stderr)
        exit(2)
    print("Lines", "Corect", "OOV", sep="\t", file=options.statfile)
    print(lines, correct, oov, sep="\t", file=options.statfile)
    print(lines / lines * 100 if lines != 0 else 0,
          correct / lines * 100 if lines != 0 else 0,
          oov / lines * 100,
          sep="\t", file=options.statfile)
    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Just a hack to hide some ugly twolc stuff in a python file."""

from sys import stderr

from ..settings import (common_multichars, fin_consonants, fin_lowercase,
                        fin_symbols, fin_uppercase, fin_vowels,
                        newword_boundary, optional_hyphen, word_boundary)
from ..string_manglers import twolc_escape


def format_copyright_twolc():
    """Generate a copyright notice in twolc format."""
    return """
! This automatically generated twolc data is originated from
! omorfi database.
! Copyright (c) 2018 Omorfi contributors

! This program is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, version 3 of the License

! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.

! You should have received a copy of the GNU General Public License
! along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""


def format_alphabet_twolc(_, ruleset):
    """Generate pair alphabet in twolc format."""
    twolcstring = 'Alphabet\n'
    if ruleset.startswith('recase'):
        twolcstring += '! Set of Finnish alphabets generated from python:\n'
        for c in fin_lowercase:
            twolcstring += c + '! allow lowercase as is\n'
            twolcstring += c + ':' + c.upper() + '! allow uppercasing\n'
        for c in fin_uppercase:
            twolcstring += c + '! allow uppercase as is\n'
            twolcstring += c + ':' + c.lower() + '! allow lowercasing\n'
        for mcs in common_multichars:
            twolcstring += twolc_escape(mcs) + '\n'
    elif ruleset.startswith('uppercase'):
        twolcstring += '! Set of Finnish alphabets generated from python:\n'
        for c in fin_lowercase:
            twolcstring += c + '! allow lowercase as is\n'
            twolcstring += c + ':' + c.upper() + '! allow uppercasing\n'
        for c in fin_uppercase:
            twolcstring += c + '! allow uppercase as is\n'
        for mcs in common_multichars:
            twolcstring += twolc_escape(mcs) + '\n'
    elif ruleset == 'hyphenate':
        twolcstring += ' '.join(fin_lowercase) + '! lower\n'
        twolcstring += ' '.join(fin_uppercase) + '! upper\n'
        for mcs in common_multichars:
            twolcstring += twolc_escape(mcs) + ':0 ! deleting all specials\n'
            if mcs == optional_hyphen or mcs == word_boundary or\
                    mcs == newword_boundary:
                twolcstring += twolc_escape(mcs) + \
                    ':%-1 ! always hyphen or nothing\n'
        twolcstring += '0:%-2 ! weaker hyphens\n'
    elif ruleset == 'hyphens':
        twolcstring += twolc_escape(optional_hyphen) + \
            ':0  ! boundary can be zero\n'
        twolcstring += twolc_escape(optional_hyphen) + \
            ':%- ! or (ASCII) hyphen\n'
        twolcstring += '%-\n'
        for mcs in common_multichars:
            if mcs != optional_hyphen:
                twolcstring += twolc_escape(mcs) + '\n'
    elif ruleset == 'apertium':
        for mcs in common_multichars:
            twolcstring += twolc_escape(mcs) + ':0 ! deleting all specials\n'
    elif ruleset == 'phon':
        for mcs in common_multichars:
            twolcstring += twolc_escape(mcs) + '\n'
    else:
        print("Unknown alphabet for ruleset", ruleset, file=stderr)
        exit(1)
    twolcstring += ';\n'
    return twolcstring


def format_sets_twolc(_, ruleset):
    """Generate alphabet subsets in twolc format."""
    twolcstring = 'Sets\n'
    if ruleset.startswith('uppercase') or ruleset.startswith('recase'):
        twolcstring += 'Lower = ' + ' '.join(fin_lowercase) + ' ;' + \
            '! Lowercase alphabets\n'
        twolcstring += 'Upper = ' + ' '.join(fin_uppercase) + ' ;' + \
            '! Uppercase alphabets\n'
    elif ruleset == 'hyphens':
        twolcstring += 'Vowels = ' + ' '.join(fin_vowels) + ' ;' + \
            '! Vowels\n'
        twolcstring += 'UpperOrSyms = ' + ' '.join(fin_uppercase) + \
            ' ' + ' '.join([twolc_escape(s) for s in fin_symbols]) +\
            '; ' + '! Symbols for likely hyphenated words\n'
    elif ruleset == 'hyphenate':
        twolcstring += 'Vowels = ' + ' '.join(fin_vowels) + ' ;' + \
            '! Vowels\n'
        twolcstring += 'Consonants = ' + ' '.join(fin_consonants) + ' ;' + \
            '! Consonants\n'
    elif ruleset == 'apertium':
        pass
    elif ruleset == 'phon':
        pass
    else:
        print("Unknown sets for ruleset", ruleset, file=stderr)
        exit(1)
    twolcstring += 'DUMMYSETCANBEUSEDTOTESTBUGS = a b c ;\n'
    return twolcstring


def format_definitions_twolc(_, ruleset):
    """Generate Regex definitions in twolc foramt."""
    twolcstring = 'Definitions\n'
    if ruleset == 'hyphenate':
        twolcstring += 'WordBoundary = [ %- | :%- | ' \
            + word_boundary + ':0 | #: | .#. ] ;\n'
    twolcstring += 'DUMMYDEFINITIONCANBEUSEDTOTESTBUGS = a | b | c ;\n'
    return twolcstring


def format_rules_twolc(_, ruleset):
    """Generate rules in twolc format."""
    twolcstring = "Rules\n"
    if ruleset == 'stub-phon' or ruleset == 'phon':
        twolcstring += '"Dummy rule"\na <= _ ;\n'
    elif ruleset == 'recase-any':
        twolcstring += '"Uppercase anywhere dummy rule"\n'
        twolcstring += twolc_escape(optional_hyphen) + " <= _ ;\n"
    elif ruleset == 'uppercase-first':
        twolcstring += '"Require uppercase in beginning"\n'
        twolcstring += 'LC:UC => .#. _ ;\n'
        twolcstring += '\twhere LC in Lower UC in Upper matched ;\n'
    elif ruleset == 'uppercase-any':
        twolcstring += '"Disallow lowercase"\n'
        twolcstring += 'UC:LC /<= _ ;\n'
        twolcstring += '\twhere LC in Lower UC in Upper matched ;\n'
    elif ruleset == 'hyphens':
        twolcstring += '"Disallow no hyphen between equal vowels"\n'
        twolcstring += twolc_escape(optional_hyphen) + ':0 /<= ' + \
            "VOWEL :0* _ :0* VOWEL ; where VOWEL in Vowels matched ;\n"
    elif ruleset == 'hyphenate':
        twolcstring += '"Hyphenate Before consonant clusters"\n'
        twolcstring += "0:%-2 <=> Vowels (Consonants) (Consonants) _" +\
                       "Consonants Vowels ;\n"
        twolcstring += '"Hyphenate between non-diphtongs"\n'
        twolcstring += "0:%-3 <=> Vx _ Vy ;\n"
        twolcstring += "\twhere Vx in (a a a a a e e e e i i i i o o o o o " +\
                       "u u u u u y y y y y ä ä ä ä ä ö ö ö ö)\n"
        twolcstring += "\t\tVy in (e o y ä ö a o ä ö a o ä ö a e y ä ö " +\
                       "a e y ä ö e ä a o u e ö a o u ä a o u) matched ;\n"
        twolcstring += '"Hyphenate diphtongs in latter syllables"\n'
        twolcstring += "0:%-4 <=> WordBoundary (Consonants) (Consonants) " +\
                       "[Vowels (Vowels) Consonants (Consonants)]+ Vx _ Vy ;\n"
        twolcstring += "\twhere Vx in (a e o u y ä ö a e i o ä ö u y i e i)\n"
        twolcstring += "\t\tVy in (i i i i i i i u u u u " +\
                       "y y o ö y y e) matched ;\n"
    elif ruleset == 'apertium':
        twolcstring += '"Remove stuffs"\n'
        twolcstring += "a <= _ ; ! remove everywhere\n"
    else:
        print("Unknown ruleset", ruleset, file=stderr)
        exit(1)
    return twolcstring



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Neat error handling for omorfi scripts"""

# Author: Tommi A Pirinen <flammie@iki.fi> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from sys import stderr


def fail_formatting_missing_for(stuff, format, moar=None):
    print("\033[93mMissing tag!\033[0m Trying to format:", stuff, "for tagset:",
          format, file=stderr)
    if moar:
        print("\033[92mExplanation\033[0m:", moar, file=stderr)


def fail_guess_because(wordmap, matches, failures, moar=None):
    print("\033[93mUnguessable!\033[0m Following has been seen:", matches,
          "\nfollowing was tested and failed:", failures,
          "\nwhen trying: (",
          wordmap['lemma'], wordmap['pos'],
          wordmap['kotus_tn'], wordmap['kotus_av'],
          wordmap['harmony'], ")", file=stderr)
    if moar:
        print("\033[92mExplanation\033[0m:", moar, file=stderr)
    # This can be used to automate classifying plurales etc.
    # if wordmap['lemma'].endswith('t'):
    # if wordmap['is_proper']:
    #    print(wordmap['lemma'], wordmap['kotus_tn'], wordmap['kotus_av'])


def just_fail(because, file=stderr):
    print("\033[93mError!\033[0m ", because, file=stderr)
